﻿using System;

namespace BinaryTreeUI
{
	//http://visualstudiogallery.msdn.microsoft.com/60297607-5fd4-4da4-97e1-3715e90c1a23/
    public class RBTree: Tree<RBTreeNode>
    {
        #region Insert 
        
        protected override void Insert(RBTreeNode node)
        {
            if (Root == null)
            {
                Root = node;
                node.IsRed = false;
            }
			else{
				node.IsRed = true;
            	FindPosition(Root, node);
            	AdjustNewRed(node);
			}
        }

        /// <summary>
        /// Adjust the tree's node location and color.
        /// </summary>
        /// <param name="node"></param>
        private void AdjustNewRed(RBTreeNode node)
        {
            var parent = node.Parent;
            if (parent == null)
            {
                node.IsRed = false; //Root
                return;
            }

            if (!parent.IsRed) return;

            var uncle = GetBrotherNode(parent);
            var grand = parent.Parent;

            if (uncle !=null && uncle.IsRed)
            {
                uncle.IsRed = parent.IsRed = false;
                grand.IsRed = true;
                AdjustNewRed(grand);
            }
            else
            {
                if (parent == grand.LeftChild)
                    RightRotateForNewRed(node, parent, grand);
                else LeftRotateForNewRed(node, parent, grand);
            }
        }

				        /// <summary>
        /// Rotate the tree, which may contain two rotatations. And change the color.
        /// </summary>
        /// <param name="node"></param>
        /// <param name="parent"></param>
        /// <param name="grand"></param>
        private void LeftRotateForNewRed(RBTreeNode node, RBTreeNode parent, RBTreeNode grand)
        {
            if (node == parent.LeftChild)
            {
                RightRotate(node, parent);
                parent = node;
            }

            LeftRotate(parent, grand);

            ReverseColor(parent, grand);
        }
		        /// <summary>
        /// Rotate the tree, which may contain two rotatations. And change the color.
        /// </summary>
        /// <param name="node"></param>
        /// <param name="parent"></param>
        /// <param name="grand"></param>
        private void RightRotateForNewRed(RBTreeNode node, RBTreeNode parent, RBTreeNode grand)
        {
            if (node == parent.RightChild)
            {
                LeftRotate(node, parent);
                parent = node;
            }

            RightRotate(parent, grand);

            ReverseColor(parent, grand);
        }
		        #endregion

        #region Common Operation

        private static void ReverseColor(RBTreeNode parent, RBTreeNode grand)
        {
            grand.IsRed = !grand.IsRed;
            parent.IsRed = !parent.IsRed;
            if (parent.Parent == null) parent.IsRed = false;    //Root
        }

        static RBTreeNode GetBrotherNode(RBTreeNode node)
        {
            var parent = node.Parent;
            var brother = node == parent.LeftChild? parent.RightChild:parent.LeftChild;
            return brother;
        }
        #endregion

        #region Remove

        protected override void Remove(RBTreeNode node)
        {
            if (node == Root)
            {
                Root = null;
                return;
            }

            var parent = node.Parent;
            var child = (node.LeftChild ?? node.RightChild);
            if (child != null) child.Parent = parent;

            var isLeft = node == parent.LeftChild;
            if (isLeft) parent.LeftChild = child;
            else parent.RightChild = child;

            if (child != null && child.IsRed)
            {
                child.IsRed = false;
            }
            else if (!node.IsRed)
            {
                node = child;
                RBTreeNode brother;
                if (isLeft)
                {
                    brother = parent.RightChild;
                    AdjustLeftMissedBlack(node, parent, brother);
                }
                else
                {
                    brother = parent.LeftChild;
                    AdjustRightMissedBlack(node, parent, brother);
                }
            }
        }

		        private void LeftRotateToAddBlack(RBTreeNode node, RBTreeNode brother, RBTreeNode parent)
        {
            LeftRotate(brother, parent);
            brother.IsRed = parent.IsRed;
            parent.IsRed = node.IsRed = false;
        }

        private void AdjustLeftMissedBlack(RBTreeNode node, RBTreeNode parent, RBTreeNode brother)
        {
            if (brother.IsRed)
            {
                LeftRotate(brother, parent);
                ReverseColor(brother, parent);
                brother = parent.RightChild;
            }

            var oppChild = brother.LeftChild;
            var child = brother.RightChild;

            if (child != null && child.IsRed)
            {
                LeftRotateToAddBlack(child, brother, parent);
            }
            else if (oppChild != null && oppChild.IsRed)
            {
                RightRotate(oppChild, brother);
                ReverseColor(oppChild, brother);
                LeftRotateToAddBlack(brother, oppChild, parent);
            }
            else
            {
                brother.IsRed = true;
                if (parent.IsRed)
                {
                    parent.IsRed = false;
                }
                else
                {
                    node = parent;
                    parent = node.Parent;
                    if (parent != null)
                    {
                        brother = GetBrotherNode(node);
                        if (node == parent.LeftChild)
                            AdjustLeftMissedBlack(node, parent, brother);
                        else AdjustRightMissedBlack(node, parent, brother);
                    }
                }
            }
        }
		        private void RightRotateToAddBlack(RBTreeNode node, RBTreeNode brother, RBTreeNode parent)
        {
            RightRotate(brother, parent);
            brother.IsRed = parent.IsRed;
            parent.IsRed = node.IsRed = false;
        }

        private void AdjustRightMissedBlack(RBTreeNode node, RBTreeNode parent, RBTreeNode brother)
        {
            if (brother.IsRed)
            {
                RightRotate(brother, parent);
                ReverseColor(brother, parent);
                brother = parent.LeftChild;
            }

            var oppChild = brother.RightChild;
            var child = brother.LeftChild;

            if (child != null && child.IsRed)
            {
                RightRotateToAddBlack(child, brother, parent);
            }
            else if (oppChild != null && oppChild.IsRed)
            {
                LeftRotate(oppChild, brother);
                ReverseColor(oppChild, brother);
                RightRotateToAddBlack(brother, oppChild, parent);
            }
            else
            {
                brother.IsRed = true;
                if (parent.IsRed)
                {
                    parent.IsRed = false;
                }
                else
                {
                    node = parent;
                    parent = node.Parent;
                    if (parent != null)
                    {
                        brother = GetBrotherNode(node);
                        if (node == parent.LeftChild)
                            AdjustLeftMissedBlack(node, parent, brother);
                        else AdjustRightMissedBlack(node, parent, brother);
                    }
                }
            }
        }
		
        #endregion
    }
}