﻿using System;
using System.Windows.Media;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Shapes;

namespace BinaryTreeUI
{

    class NodePainter<T> where T:TreeNode<T,int>
    {
        NodeShapeOptions options = new NodeShapeOptions();
        Canvas panel;

        public void Paint(T root, Canvas panel)
        {
            this.panel = panel;
            var area = this.CreateNodeShape(root);

            //Middle top, a little margin to top border
            area.Margin = new Thickness((panel.ActualWidth - options.Diameter) / 2, 10, 0, 0);
            panel.Children.Add(area);
            PaintChild(root, area.Margin);
            AdjustLayout();
        }

        private void AdjustLayout()
        {
            panel.UpdateLayout();

            foreach (UIElement item in panel.Children)
            {
                var area = item as Canvas;
                if (area == null) continue;
                var orb = area.Children[0] as FrameworkElement;     //Ellispe
                var tb = area.Children[1] as FrameworkElement;      //TextBlock
                var left = orb.Width / 2 - tb.ActualWidth / 2;
                var top = orb.Height / 2 - tb.ActualHeight / 2;
                tb.Margin = new Thickness(left, top, 0, 0);
            }
        }

        void PaintChild(T parent, Thickness margin, int level = 1)
        {
            var diff = new Size(options.Diff.Width * Math.Pow(0.55, level - 1), options.Diff.Height * Math.Pow(0.9, level - 1));
            var chord = Math.Sqrt(diff.Width * diff.Width + diff.Height * diff.Height);

            //The distance from center of Circle to the line end.
            var chord2 = options.Diameter / 2 + options.LineFromShape;
            double lineDiffX = chord2 * diff.Width / chord;
            double lineDiffY = chord2 * diff.Height / chord;

            if (parent.LeftChild != null)
            {
                var area = this.CreateNodeShape(parent.LeftChild);
                area.Margin = new Thickness(margin.Left - diff.Width, margin.Top + diff.Height, 0, 0);
                panel.Children.Add(area);

                var line1 = new Line
                {
                    Stroke = options.LineStroke,
                    StrokeThickness = options.LineThickness,
                    X1 = margin.Left + options.Diameter / 2 - lineDiffX,
                    Y1 = margin.Top + options.Diameter / 2 + lineDiffY,
                    X2 = area.Margin.Left + options.Diameter / 2 + lineDiffX,
                    Y2 = area.Margin.Top + options.Diameter / 2 - lineDiffY
                };
                panel.Children.Add(line1);
                PaintChild(parent.LeftChild, area.Margin, level +1);
            }

            if (parent.RightChild != null)
            {
                var area = this.CreateNodeShape(parent.RightChild);
                area.Margin = new Thickness(margin.Left + diff.Width, margin.Top + diff.Height, 0, 0);
                panel.Children.Add(area);

                var line1 = new Line
                {
                    Stroke = options.LineStroke,
                    StrokeThickness = options.LineThickness,
                    X1 = margin.Left + options.Diameter / 2 + lineDiffX,
                    Y1 = margin.Top + options.Diameter / 2 + lineDiffY,
                    X2 = area.Margin.Left + options.Diameter / 2 - lineDiffX,
                    Y2 = area.Margin.Top + options.Diameter / 2 - lineDiffY
                };
                panel.Children.Add(line1);
                PaintChild(parent.RightChild, area.Margin, level+1);
            }
        }

        Canvas CreateNodeShape(T node)
        {
            var area = new Canvas();
            var orb = new Ellipse
            {
                Width = options.Diameter,
                Height = options.Diameter,
                Stroke = options.Stroke,
                Fill = options.Fill
            };
            area.Children.Add(orb);

            var tb = new TextBlock { Text = node.Value.ToString(), FontSize = options.FontSize };
            area.Children.Add(tb);

            SetColorForIfRTNode(node, orb, tb);
            return area;
        }

        private void SetColorForIfRTNode(T node, Ellipse orb, TextBlock tb)
        {
            var rtNode = node as RBTreeNode;
            if (rtNode != null && !rtNode.IsRed)
            {
                orb.Fill = Brushes.Green;
                tb.Foreground = Brushes.White;
            }

            var avlNode = node as AvlTreeNode;
            if (avlNode != null)
            {
                orb.ToolTip = avlNode.ToString();
            }
        }

    }

    class NodeShapeOptions
    {
        public double LineFromShape { get; set; }

        public Brush Stroke { get; set; }

        public Brush Fill { get; set; }

        public double Diameter { get; set; }

        public int FontSize { get; set; }

        public Brush ForeGround { get; set; }

        public Size Diff { get; set; }

        public int LineThickness { get; set; }

        public Brush LineStroke { get; set; }

        public NodeShapeOptions()
        {
            this.FontSize = 14;
            this.Diameter = 50;
            this.Fill = Brushes.Coral;
            this.Stroke = Brushes.Black;
            this.ForeGround = Brushes.Black;
            this.Diff = new Size(300, 150);
            this.LineStroke = Brushes.Blue;
            this.LineThickness = 3;
            this.LineFromShape = 5;
        }
    }

    /// <summary>
    /// 
    /// </summary>
    /// <typeparam name="T">Type of the node.</typeparam>
    /// <typeparam name="K">Type of the node value.</typeparam>
    public class TreeNode<T,K> where T:TreeNode<T,K> where K: IComparable<K>
    {
        public T LeftChild { get; set; }

        public T RightChild { get; set; }

        public T Parent { get; set; }

        public K Value { get; set; }
    }

    class TreeNode : TreeNode<TreeNode, Int32> { }

    public class RBTreeNode : TreeNode<RBTreeNode,Int32 >
    {
        /// <summary>
        /// 树节点颜色，是否为红。
        /// </summary>
        public bool IsRed { get; set; }

        public override string ToString()
        {
            return this.Value + "," + (this.IsRed ? "R" : "B");
        }
    }

    public class AvlTreeNode : TreeNode<AvlTreeNode,Int32>
    {
        /// <summary>
        /// 节点的平衡度
        /// </summary>
        public int Balance { get; set; }

        public override string ToString()
        {
            return "Balance: " + Balance + ", Value: " + this.Value;
        }
    }

    public abstract class Tree<T> where T:TreeNode<T,Int32>, new()
    {
        public T Root { get; protected set; }

        /// <summary>
        /// Find a new node's position under a given root.
        /// </summary>
        protected void FindPosition(T root, T node)
        {
            if (node.Value < root.Value)
            {
                if (root.LeftChild == null)
                {
                    root.LeftChild = node;
                    node.Parent = root;
                }
                else FindPosition(root.LeftChild, node);
            }
            else
            {
                if (root.RightChild == null)
                {
                    root.RightChild = node;
                    node.Parent = root;
                }
                else FindPosition(root.RightChild, node);
            }
        }

        T FindMax(T root)
        {
            while (root.RightChild != null)
            {
                root = root.RightChild;
            }
            return root;
        }

        T FindMin(T root)
        {
            while (root.LeftChild != null)
            {
                root = root.LeftChild;
            }
            return root;
        }

        #region Rotation   
        protected void LeftRotate(T node, T parent)
        {
            OperateForBothRotation(node, parent);

            parent.RightChild = node.LeftChild;
            if (node.LeftChild != null)
            {
                node.LeftChild.Parent = parent;
            }
            node.LeftChild = parent;
        }

        protected void RightRotate(T node, T parent)
        {
            OperateForBothRotation(node, parent);

            parent.LeftChild = node.RightChild;
            if (node.RightChild != null)
            {
                node.RightChild.Parent = parent;
            }
            node.RightChild = parent;
        }

        private void OperateForBothRotation(T node, T parent)
        {
            var grand = parent.Parent;

            node.Parent = grand;
            parent.Parent = node;

            if (grand == null)
            {
                Root = node;
            }
            else if (parent == grand.RightChild) grand.RightChild = node;
            else grand.LeftChild = node;
        }
        #endregion

        #region Remove

        public bool Remove(int value)
        {
            T root = this.Root;
            while (root != null)
            {
                if (value == root.Value)
                {
                    T node = root;
                    if (root.LeftChild != null) node = FindMax(root.LeftChild);
                    else if (root.RightChild != null) node = FindMin(root.RightChild);

                    root.Value = node.Value;
                    this.Remove(node);      //子类的调整旋转操作.

                    return true;
                }
                else if (value < root.Value) root = root.LeftChild;
                else root = root.RightChild;
            }

            return false;
        }

        protected abstract void Remove(T node);

        #endregion

        #region Insert

        public void Insert(int value)
        {
            var node = new T { Value = value };
            this.Insert(node);
        }

        protected abstract void Insert(T node);
        #endregion
    }

    class OrdinaryTree : Tree<TreeNode>
    {
        protected override void Remove(TreeNode node)
        {
            var parent = node.Parent;

            if (parent == null)
            {
                this.Root = null;
            }
            else if (node == parent.LeftChild)
            {
                parent.LeftChild = null;
            }
            else parent.RightChild = null;
        }

        protected override void Insert(TreeNode node)
        {
            if (Root == null)
                Root = node;
            else 
                FindPosition(Root, node);
        }
    }
}
